/*
 * Copyright 2008 Carlo Bramini
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#define _WIN32_WINNT    0x0500

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#ifdef __MINGW32__
#include <ntdef.h>
#else
#include <winternl.h>
#endif

#include <stdarg.h>
#include <math.h>
#include <float.h>
#include <ctype.h>
#include <sys/types.h>

/*
#define FILE        void
#define EOF         -1
#define stdout      NULL
*/
#define ptrdiff_t   int

#ifdef __GNUC__
#define INT64_K(_k) _k##LL
#else
#define INT64_K(_k) (__int64)_k
#endif

#define SIZEOF(_v)  (sizeof(_v)/sizeof(*_v))

#define _MK(_c0, _c1, _c2) \
    (((_c0) << 16) | ((_c1) << 8) | (_c2))

/* Define if you want to use long long (8 byte) qualifier:
   http://msdn.microsoft.com/en-us/library/s3f49ktz(VS.80).aspx
*/
/*#define HAVE_LONG_LONG*/

/* Define this macro if sizeof(long double) == sizeof(double) */
#define LONG_DOUBLE_IS_DOUBLE   1

#define FLG_LEFT_ALIGN      0x001
#define FLG_INCLUDE_SIGN    0x002
#define FLG_SPACE_IF_PLUS   0x004
#define FLG_SPECIAL         0x008
#define FLG_PAD_WITH_0      0x010
#define FLG_IS_SIGNED       0x020
#define FLG_IS_ZERO         0x040
#define FLG_CHECK_SIGN      0x080
#define FLG_IS_GFLAG        0x100
#define FLG_SCI_MODE        0x200

#define QL_h    _MK(0, 0, 'h')
#define QL_l    _MK(0, 0, 'l')
#define QL_ll   _MK(0, 'l', 'l')
#define QL_L    _MK(0, 0, 'L')
#define QL_w    _MK(0, 0, 'w')
#define QL_I    _MK(0, 0, 'I')
#define QL_I32  _MK('I', '3', '2')
#define QL_I64  _MK('I', '6', '4')

#ifdef LONG_DOUBLE_IS_DOUBLE
#define N_DIGIT_EXPONENT    3
#else
#define N_DIGIT_EXPONENT    4
#endif

typedef BOOL (*write_char_t)(void *descriptor, int character);
typedef int  (*read_char_t)(void *descriptor, int increments);

typedef struct {
    unsigned int prefix;
    unsigned int width;
    signed int   precision;
    unsigned int size;
    int          type;
    int          sign;
} stack_escape_t;

typedef struct {
    read_char_t  pfin;
    void        *ds_in;
    write_char_t pfout;
    void        *ds_out;
    BOOL         is_unicode;
} driver_t;

#define PFIN(_n)    drv->pfin(&drv->ds_in, _n)
#define PFOUT(ch)   drv->pfout(drv->ds_out, ch)

static const char digits[16] = "0123456789abcdef";
static char decimal = '.';

int asc2number(int code)
{
    char *ptr;

    code = tolower(code);
    ptr = strchr(digits, code);
    if (ptr != NULL)
        return ptr - digits;
    return 0;
}

#define READ_TYPED(_type, _mask) \
    if (esc->prefix & FLG_CHECK_SIGN) \
        num=va_arg(*args, signed _type); \
    else \
        num=va_arg(*args, unsigned _type) & _mask

#ifdef __GNUC__
#define MASK_SHORT  (1ULL | (((1ULL << (sizeof(short)*8 - 1))-1) << 1))
#define MASK_LONG   (1ULL | (((1ULL << (sizeof(long)*8 - 1))-1) << 1))
#define MASK_INT32  (1ULL | (((1ULL << (sizeof(__int32)*8 - 1))-1) << 1))
#define MASK_INT64  (1ULL | (((1ULL << (sizeof(__int64)*8 - 1))-1) << 1))
#define MASK_INT    (1ULL | (((1ULL << (sizeof(int)*8 - 1))-1) << 1))
#else
#define MASK_SHORT  ((unsigned __int64)1 | (((unsigned __int64)1 << (sizeof(short)*8 - 1))-1) << 1)
#define MASK_LONG   ((unsigned __int64)1 | (((unsigned __int64)1 << (sizeof(long)*8 - 1))-1) << 1)
#define MASK_INT32  ((unsigned __int64)1 | (((unsigned __int64)1 << (sizeof(__int32)*8 - 1))-1) << 1)
#define MASK_INT64  ((unsigned __int64)1 | (((unsigned __int64)1 << (sizeof(__int64)*8 - 1))-1) << 1)
#define MASK_INT    ((unsigned __int64)1 | (((unsigned __int64)1 << (sizeof(int)*8 - 1))-1) << 1)
#endif

static int fill_char(driver_t *drv, stack_escape_t *esc, unsigned int start, int fill)
{
    while (start < esc->width)
    {
        if (!PFOUT(fill))
            break;
        start++;
    }
    return start;
}

static int write_number(driver_t *drv, stack_escape_t *esc, va_list *args, int base)
{
    unsigned __int64 num;
    char   tmp[80];
    unsigned int idx, rem;

    /* Read parameter according to specifications */
    switch (esc->size) {
    case QL_h:   READ_TYPED(int, MASK_SHORT); break;
    case QL_l:   READ_TYPED(long, MASK_LONG); break;
    case QL_I32: READ_TYPED(__int32, MASK_INT32); break;
#ifdef HAVE_LONG_LONG
    case QL_ll:
#endif
    case QL_I64: READ_TYPED(__int64, MASK_INT64); break;
    case QL_I:
        if (esc->type == 'd' || esc->type == 'i')
            num = va_arg(*args, ptrdiff_t);
        else
            num = va_arg(*args, size_t);
        break;

    default:
        READ_TYPED(int, MASK_INT); break;
    }

    /* Some types are unsigned by default,
     * so we do not handle their sign */
    if ((esc->prefix & FLG_CHECK_SIGN))
    {
        /* Otherwise, check if sign bit is 1 */
        if ((signed __int64)num < 0)
        {
            /* we need to work with unsigned types */
            esc->prefix |= FLG_IS_SIGNED;
            num = 0 - num;
        }
    }

    /* we use this flag to avoid negative zero */
    if (num == 0)
        esc->prefix |= FLG_IS_ZERO;

    /* expand the digits of the number */
    idx = 0;
    do
    {
        rem = (unsigned int)(num % base);
        num = num / base;
        tmp[idx++] = digits[rem];
    } while (num != 0);

    /* Check prefixes for hex types */
    if (base == 16)
    {
        switch (esc->prefix & (FLG_PAD_WITH_0 | FLG_SPECIAL)) {
        case FLG_PAD_WITH_0:
            while (idx < esc->width)
                tmp[idx++] = '0';
            break;
        case FLG_PAD_WITH_0 | FLG_SPECIAL:
            while ((idx+2) < esc->width)
                tmp[idx++] = '0';
        case FLG_SPECIAL:
            if (!(esc->prefix & FLG_IS_ZERO))
            {
                tmp[idx++] = 'x';
                tmp[idx++] = '0';
            }
            break;
        }
        /* test if number must be capitalized */
        if (esc->type == 'X')
            _strupr(tmp);
    } else
    if (base == 8)
    {
        if ((esc->prefix & (FLG_SPECIAL | FLG_IS_ZERO)) == FLG_SPECIAL)
            tmp[idx++] = '0';
        if ((esc->prefix & FLG_PAD_WITH_0))
        {
            while (idx < esc->width)
                tmp[idx++] = '0';
        }
    } else {
        /* base 10 */
        if ((esc->prefix & FLG_IS_SIGNED))
        {
            if ((esc->prefix & FLG_PAD_WITH_0))
            {
                while ((idx+1) < esc->width)
                    tmp[idx++] = '0';
            }
            tmp[idx++] = '-';
        } else {
            if ((esc->prefix & FLG_PAD_WITH_0))
            {
                while (idx < esc->width)
                    tmp[idx++] = '0';
            }
            if ((esc->prefix & FLG_INCLUDE_SIGN))
            {
                tmp[idx++] = '+';
            } else
            if ((esc->prefix & FLG_SPACE_IF_PLUS))
            {
                tmp[idx++] = ' ';
            }
        }
    }
    /* number is finished (pheeuu...) */
    tmp[idx] = '\0';

    /* preserve the size of the number */
    rem = idx;

    /* check if the number must be aligned to right */
    if (!(esc->prefix & FLG_LEFT_ALIGN))
    {
        rem = fill_char(drv, esc, rem, ' ');
    }

    /* output the number */
    while (idx > 0)
    {
        if (!PFOUT(tmp[--idx]))
            break;
    }

    /* check if the number must be aligned to left */
    if ((esc->prefix & FLG_LEFT_ALIGN))
    {
        rem = fill_char(drv, esc, rem, ' ');
    }

    /* number of emitted chars */
    return rem;
}

static int write_string_a(driver_t *drv, stack_escape_t *esc, const char *stra, unsigned int len)
{
    unsigned int fill;

    /* if width is specified into precision,
     * we must move the values into the right places */
    if (esc->width == 0 && esc->precision != -1)
        esc->width = esc->precision;

    /* check if the string really exists */
    if (stra == NULL) {
        stra = "(null)";
        len = 6;
    }

    /* string may be padded with spaces or zeros */
    if ((esc->prefix & FLG_PAD_WITH_0))
        fill = '0';
    else
        fill = ' ';

    /* check if the string must be aligned to right */
    if (!(esc->prefix & FLG_LEFT_ALIGN))
        len = fill_char(drv, esc, len, fill);

    /* output the string */
    while (*stra)
    {
        if (!PFOUT(*stra++))
            break;
    }

    /* check if the string must be aligned to left */
    if ((esc->prefix & FLG_LEFT_ALIGN))
        len = fill_char(drv, esc, len, ' ');

    return len;
}

static int write_string_w(driver_t *drv, stack_escape_t *esc, const wchar_t *strw, unsigned int len)
{
    unsigned int fill;

    /* if width is specified into precision,
     * we must move the values into the right places */
    if (esc->width == 0 && esc->precision != -1)
        esc->width = esc->precision;

    /* check if the string really exists */
    if (strw == NULL) {
        strw = L"(null)";
        len = 6;
    }

    /* string may be padded with spaces or zeros */
    if ((esc->prefix & FLG_PAD_WITH_0))
        fill = '0';
    else
        fill = ' ';

    /* check if the string must be aligned to right */
    if (!(esc->prefix & FLG_LEFT_ALIGN))
        len = fill_char(drv, esc, len, fill);

    /* output the string */
    while (*strw)
    {
        if (!PFOUT(*strw++))
            break;
    }

    /* check if the string must be aligned to left */
    if ((esc->prefix & FLG_LEFT_ALIGN))
        len = fill_char(drv, esc, len, ' ');

    return len;
}

static void write_n_processed(driver_t *drv, stack_escape_t *esc, int count, void *argp)
{
    union {
        void               *ptr_void;
        unsigned int       *ptr_int;
        unsigned long      *ptr_long;
        unsigned __int32   *ptr_int32;
        unsigned __int64   *ptr_int64;
        unsigned short int *ptr_short;
        size_t             *ptr_size;
    } n_out;

    /* should be easy... */
    n_out.ptr_void = argp;

    /* Read parameter according to specifications */
    switch (esc->size) {
    case QL_h:   *n_out.ptr_short = count; break;
    case QL_l:   *n_out.ptr_long  = count; break;
    case QL_I32: *n_out.ptr_int32 = count; break;
#ifdef HAVE_LONG_LONG
    case QL_ll:
#endif
    case QL_I64: *n_out.ptr_int64 = count; break;
    case QL_I:   *n_out.ptr_size  = count; break;

    default:     *n_out.ptr_int   = count; break;
    }
}

static long double acquire_number(stack_escape_t *esc, va_list *args)
{
    long double res;

    /* Read parameter according to specifications */
    switch (esc->size) {
#ifndef LONG_DOUBLE_IS_DOUBLE
    case QL_L: res = va_arg(*args, long double); break;
#endif
    case QL_l: res = va_arg(*args, double); break;
    default:   res = va_arg(*args, double); break;
    }
    return res;
}

static const char adder_tbl[11] = { '0','1','2','3','4','5','6','7','8','9','0' };
static const char carry_tbl[11] = {   0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  1 };

static unsigned int unpack_number(long double number, stack_escape_t *esc, char *tmp, unsigned int size)
{
    long double exponent;
    signed long int iexponent = 0;
    char wrk[256];
    char *swrk, *fwrk, *pwrk, *dwrk;
    char digit, carry;

    /* if the number is negative, make
     * it positive and set the flag */
    if (number < 0)
    {
        number = 0 - number;
        /* we need to work with unsigned types */
        esc->prefix |= FLG_IS_SIGNED;
    }

    /* leave the space for sign */
    esc->sign = 0;
    if ((esc->prefix & FLG_IS_SIGNED))
    {
        esc->sign = '-';
    } else {
        if ((esc->prefix & FLG_INCLUDE_SIGN))
        {
            esc->sign = '+';
        } else
        if ((esc->prefix & FLG_SPACE_IF_PLUS))
        {
            esc->sign = ' ';
        }
    }

    /* try to discover desidered exponent */
    if (number != 0)
    {
        /* get exponent of normalized number */
        modfl(log10(number), &exponent);
        /* if integer part of the number is zero, 
         * adjust the exponent
         */
        if (number < 1)
            exponent -= 1;
        iexponent = (signed long int)(exponent);
    }
    /* transform the number to be in scientific
     * notation (easier to handle)
     */
    number *= pow(10., 0-iexponent);

    /* precision undefined? assume 6 digits */
    if (esc->precision < 0)
        esc->precision = 6;

    /* check if the user has selected scientific mode.
     * From now, the "prefix" member will keep this
     * selection too so we can forgot "type" member.
     */
    if (esc->type == 'e' || esc->type == 'E')
        esc->prefix |= FLG_SCI_MODE;

    /* for gG formats, decide here the switch between eE and fF */
    if (esc->type == 'g' || esc->type == 'G')
    {
        /* ANSI */
        if (esc->precision == 0)
            esc->precision = 1;

        /* check if we need scientific mode */
        if (iexponent >= -4 && iexponent < esc->precision)
            esc->precision -= iexponent;
        else {
            esc->prefix |= FLG_SCI_MODE;
            esc->type -= 2;
        }

        /* gG outputs DOT only if there are some significant digits. */
        if (number != 0)
            esc->precision--;

        /* remove all trailing, non significant zeros */
        if (!(esc->prefix & FLG_SPECIAL))
            esc->prefix |= (FLG_SPECIAL | FLG_IS_GFLAG);
    }
    /* setup the point where integer part will be emitted.
     * Leave one buck free if rounding will occur.
     */
    swrk = wrk + 1;
    pwrk = swrk;

    /* SCIENTIFIC mode? Then limit integer output */
    if ((esc->prefix & FLG_SCI_MODE))
        dwrk = swrk;
    else {
        /* if integer part is zero, we emit the obvious zero and the dot */
        if (iexponent < 0)
        {
            *pwrk++ = '0';
            /* if the user forced some digits after dot or
             * dor output is forced by special flag, we must
             * really emit it.
             */
            if (esc->precision || (esc->prefix & FLG_SPECIAL))
                *pwrk++ = '.';

            /* calculate max number of zero to output */
            fwrk = pwrk - iexponent - 1;
            if ((fwrk-pwrk) > esc->precision) {
                fwrk = pwrk + esc->precision - 1;
                esc->precision = 0;
            } else
                esc->precision -= (fwrk-pwrk);

            /* output all zeros we need */
            while (pwrk < fwrk)
                *pwrk++ = '0';

            /* tell us that dot has been already emitted */
            dwrk = NULL;
            /* calculate here the final point since dwrk is broken */
            fwrk = pwrk + esc->precision - 1;
        } else
            dwrk = pwrk + iexponent;
    }

    /* calculates the final point except when dot is already
     * emitted because the fwrk is already calculated too.
     */
    if (dwrk != NULL)
    {
        fwrk = dwrk + esc->precision;
        /* do we really need the decimal dot? */
        if (!(esc->prefix & FLG_SPECIAL) && esc->precision == 0)
            dwrk = NULL;
        else
            fwrk++;
    }

    while (pwrk <= fwrk) {
        /* get integer part of the number
         * and advance to the next digit.
         * PS: I'm re-using 'exponent' variable
         */
        number = modfl(number, &exponent) * 10;
        *pwrk = (int)exponent + '0';
        /* check if we reached the dot point */
        if (pwrk == dwrk)
            *++pwrk = '.';
        pwrk++;
    }
    /* be sure that final point is the same of current point
     * (not guaranteed if lopp exited after emitting the dot)
     */
    fwrk = pwrk;

    /* check if rounding is required */
    if (number >= 5)
    {
        carry = 1;

        while (pwrk > swrk) {
            digit = *--pwrk;
            if (digit == '.')
                continue;
            digit = digit - '0' + carry;
            *pwrk = adder_tbl[digit];
            carry = carry_tbl[digit];
        }

        /* Output the new digit in the carry if any */
        if (carry)
        {
            /* easy addition thanks to free buck previousy reserved */
            *--swrk = '1';
            /* if scientific notation is enabled, the absolute
             * value of iexponent must be incremented too.
             */
            if ((esc->prefix & FLG_SCI_MODE))
            {
                if (iexponent > 0)
                    iexponent++;
                else
                    iexponent--;
            }
        }
    }

    /* With 'G' or 'g' types, truncate leading zeros to
     * least significant digit, but only if '#' flag is
     * not included.
     */
    if ((esc->prefix & FLG_IS_GFLAG))
    {
        while (*(fwrk - 1) == '0')
            fwrk--;
        /* also remove the dot if it's encountered and
         * there are no significant digits after it.
         */
        if (*(fwrk - 1) == '.')
            fwrk--;
    }

    /* emit the exponent */
    if ((esc->prefix & FLG_SCI_MODE))
    {
        /* output 'E' or 'e' */
        *fwrk++ = esc->type;

        /* the sign of the exponent */
        if (iexponent >= 0)
            *fwrk++ = '+';
        else {
            *fwrk++ = '-';
            iexponent = 0-iexponent;
        }
        /* the exponent itself */
        pwrk = fwrk + N_DIGIT_EXPONENT;
        while (pwrk > fwrk)
        {
            *--pwrk = '0' + (iexponent % 10);
            iexponent /= 10;
        }
        fwrk += N_DIGIT_EXPONENT;
    }

    /* FINALLY COPY PROCESSED NUMBER INTO OUTPUT BUFFER */
    memcpy(tmp, swrk, fwrk-swrk);

    return fwrk-swrk;
}

static int flush_number_fp(driver_t *drv, stack_escape_t *esc, const char *tmp, unsigned int idx)
{
    /* preserve the size of the number */
    unsigned int rem = idx, x;

    /* output '+', '-', or ' ' if exists.
     * If the number is spaced padded to right,
     * the sign must be output later but if
     * it is zeroed padded, the sign must be outputted now
     * (avoid condition like "0-1e+002")
     */
    if ((esc->sign))
    {
        /* add space used by sign */
        rem++;
        /* output the sign if it exists */
        if ((esc->prefix & FLG_PAD_WITH_0))
        {
            if (!PFOUT(esc->sign))
                return 0;
            /* sign has been emitted, so clear the field */
            esc->sign = 0;
        }
    }

    /* check if the number must be aligned to right */
    if (!(esc->prefix & FLG_LEFT_ALIGN))
    {
        /* string may be padded with spaces or zeros */
        if ((esc->prefix & FLG_PAD_WITH_0))
            rem = fill_char(drv, esc, rem, '0');
        else
            rem = fill_char(drv, esc, rem, ' ');
    }

    /* emit sign if not done yet */
    if ((esc->sign))
    {
        if (!PFOUT(esc->sign))
            return 0;
    }

    /* output the number */
    for (x=0; x<idx; x++)
    {
        if (!PFOUT(tmp[x]))
            break;
    }

    /* check if the number must be aligned to left */
    if ((esc->prefix & FLG_LEFT_ALIGN))
        rem = fill_char(drv, esc, rem, ' ');

    /* number of emitted chars */
    return rem;
}

static int check_if_number(driver_t *drv, stack_escape_t *esc, long double number)
{
    const char *emit = NULL;

    if (_isnan(number))
        emit = "NaN";
    else
    {
        int cls = _fpclass(number);

        if (cls & _FPCLASS_NINF)
            emit = "-Inf";
        else
        if (cls & _FPCLASS_PINF)
            emit = "+Inf";
    }

    if (emit != NULL)
        return flush_number_fp(drv, esc, emit, strlen(emit));

    return 0;
}

static int write_number_fp(driver_t *drv, stack_escape_t *esc, va_list *args)
{
    int ret;
    long double number;
    char tmp[256];

    number = acquire_number(esc, args);
    ret = check_if_number(drv, esc, number);
    if (ret) return ret;

    ret = unpack_number(number, esc, tmp, SIZEOF(tmp));

    if (ret)
        ret = flush_number_fp(drv, esc, tmp, ret);

    return ret;
}

static int __rintf (driver_t *drv, va_list args)
{
    unsigned int count, ch;
    stack_escape_t esc;

    /* check if format pattern is NULL */
    if (drv->ds_in == NULL)
    {
        /* TODO: add proper stuff here! */
        return -1;
    }

    count = 0;
    while (1)
    {
        /* get a character */
        ch = PFIN(1);
        /* if the format string is finished, then quit */
        if (ch == '\0')
            break;
        /* if it is not an escape sequence, or it is
         * a "%%", try to output the character */
        if (ch != '%' || (ch == '%' && PFIN(0) == '%'))
        {
            /* if something went wrong, it exits */
            if (!PFOUT(ch))
                break;
            /* increase character count and restart */
            count++;
            continue;
        }
        /* detect the prefix (if any) */
        esc.prefix = 0;
        do {
            ch = PFIN(1);
            if (ch == '-') esc.prefix |= FLG_LEFT_ALIGN;
            else
            if (ch == '+') esc.prefix |= FLG_INCLUDE_SIGN;
            else
            if (ch == ' ') esc.prefix |= FLG_SPACE_IF_PLUS;
            else
            if (ch == '#') esc.prefix |= FLG_SPECIAL;
            else
            if (ch == '0') esc.prefix |= FLG_PAD_WITH_0;
            else
                break;
        } while (1);

        /* test for explicit width */
        if (ch == '*')
        {
            esc.width = va_arg(args, int);
            ch = PFIN(1);
        } else {
            /* otherwise, read until it's a number */
            esc.width = 0;
            while (isdigit(ch))
            {
                esc.width = esc.width * 10 + asc2number(ch);
                ch = PFIN(1);
            }
        }
        /* test if tere is a precision field */
        esc.precision = -1;
        if (ch == '.')
        {
            ch = PFIN(1);
            /* test for explicit precision */
            if (ch == '*') {
                esc.precision = va_arg(args, int);
                ch = PFIN(1);
            } else {
                esc.precision = 0;
                /* test if a precision field truly exists after the dot */
                if (isdigit(ch))
                {
                    do
                    {
                        esc.precision = esc.precision * 10 + asc2number(ch);
                        ch = PFIN(1);
                    } while (isdigit(ch));
                }
            }
        }
        /* check if a recognized size qualifier exists */
        esc.size = 0;
        switch (ch) {
        case 'h':   /* short */
        case 'l':   /* long */
        case 'L':   /* long double? */
        case 'w':   /* wide char/wide string */
        case 'I':   /* __int32/__int64 */
            esc.size = ch;
            ch = PFIN(1);
            break;
        }
        /* check if more processing is required for size */
        if (esc.size == 'l' && ch == 'l')
        {
            esc.size = QL_ll;
            ch = PFIN(1);
        } else
        if (esc.size == 'I')
        {
            unsigned int cl = PFIN(0);
            unsigned int tmp = _MK('I', ch, cl);

            if (tmp == QL_I32 || tmp == QL_I64)
            {
                /* discard 'cl' */
                PFIN(1);
                esc.size = tmp;
                /* get the type */
                ch = PFIN(1);
            }
        }
        /* process the modifiers */
        esc.type = ch;
        switch (ch)
        {
        case 'a':
        case 'A':
            break;
        case 'c':
        case 'C':
            if (esc.size == 'l' || esc.size == 'w' ||
                (drv->is_unicode == 0 && ch == 'C') ||
                (drv->is_unicode == 1 && ch == 'c'))
            {
                wchar_t strw[2];
                /* print unicode string */
                strw[0] = (wchar_t)va_arg(args, int);
                count += write_string_w(drv, &esc, strw, 1);
            } else {
                char stra[2];
                /* print ascii string */
                stra[0] = (unsigned char)va_arg(args, int);
                count += write_string_a(drv, &esc, stra, 1);
            }
            break;

        case 'i':
        case 'd': esc.prefix |= FLG_CHECK_SIGN;
                  count += write_number(drv, &esc, &args, 10); break;
        case 'u': esc.prefix &= ~(FLG_INCLUDE_SIGN | FLG_SPACE_IF_PLUS);
                  count += write_number(drv, &esc, &args, 10); break;
        case 'o': count += write_number(drv, &esc, &args, 8); break;
        case 'x':
        case 'X': count += write_number(drv, &esc, &args, 16); break;
        case 'e':
        case 'E':
        case 'f':
        case 'g':
        case 'G': count += write_number_fp(drv, &esc, &args); break;
        case 'n': write_n_processed(drv, &esc, count, va_arg(args, void *)); break;
        case 'p':
            break;
        case 's':
        case 'S':
            if (esc.size == 'l' || esc.size == 'w' ||
                (drv->is_unicode == 0 && ch == 'S' && esc.size != 'h') ||
                (drv->is_unicode == 1 && ch == 's' && esc.size != 'h'))
            {
                wchar_t *pu = va_arg(args, wchar_t *);
                /* print unicode string */
                count += write_string_w(drv, &esc, pu, pu ? wcslen(pu) : 0);
            } else {
                char *pa = va_arg(args, char *);
                /* print ascii string */
                /* "h" modifier always fall here! */
                count += write_string_a(drv, &esc, pa, pa ? strlen(pa) : 0);
            }
            break;
        case 'Z':
            if (esc.size == 'l' || esc.size == 'w' || drv->is_unicode) {
                /* print counted unicode string */
                PUNICODE_STRING pu = va_arg(args, PUNICODE_STRING);
                count += write_string_w(drv, &esc, pu->Buffer, pu->Length);
            } else {
                /* print counted ascii string */
                PANSI_STRING    pa = va_arg(args, PANSI_STRING);
                count += write_string_a(drv, &esc, pa->Buffer, pa->Length);
            }
            break;
        }
    }
    return count;
}

/* INPUT DIVERS */

static int achar_input(void **data_in, int add_count)
{
    char *format = *((char **)data_in);
    int   ch = *format;

    format += add_count;
    *((char **)data_in) = format;
    return ch;
}

/* OUTPUT DRIVER FOR STRINGS */
typedef struct {
    LPSTR buffer;
    UINT  size;
} astring_t;

static BOOL achar_output(astring_t *ds, int ch)
{
    if (ch == '\n' && ds->size)
        *ds->buffer++ = '\r';

    *ds->buffer++ = ch;

    return TRUE;
}

int __cdecl
my_vsprintf(LPSTR str, LPCSTR fmt, va_list argptr)
{
    int       ret;
    astring_t astr = { str, 0 };
    driver_t  drv = {
        (read_char_t) achar_input,
        (void *)      fmt,
        (write_char_t)achar_output,
        (void *)      &astr,
        FALSE
    };

    ret = __rintf(&drv, argptr);
    if (ret >= 0)
        str[ret] = 0;

    return ret;
}

int __cdecl
my_sprintf(LPSTR str, LPCSTR fmt, ...)
{
    va_list argptr;

    va_start(argptr, fmt);
    return my_vsprintf(str, fmt, argptr);
}

#include <stdio.h>
#include <conio.h>
#include <malloc.h>

void test(char *buf1, char *buf2, const char *fmt, ...)
{
    int       ret1, ret2;
    va_list   argptr;

    va_start(argptr, fmt);

    ret1 = my_vsprintf(buf1, fmt, argptr);
    ret2 = vsprintf(buf2, fmt, argptr);

    if (strcmp(buf1, buf2) || ret1 != ret2)
        printf("[ERR] %s ret=%d ret2=%d\n"
               "    calculated -> %s\n"
               "    expected   -> %s\n",
               fmt, ret1, ret2, buf1, buf2);
    else
        printf("[OK]  %s -> %s\n", fmt, buf1);
}

/*
    ESCAPE FORMAT:
    %[flag][width][len][type]
*/
void run_int_test(__int64 x)
{
    const char *str_flags[] = {
        "", "-", "+", " ", "#", "0",
    };
    const char *str_width[] = {
        "", "6", "*",
    };
    const char *str_len[] = {
        "", "h", "l", "ll", "I", "I32", "I64"
    };
    const char *str_type[] = {
        "d", "i", "o", "u", "x", "X"
    };

    char buf1[256];
    char buf2[256];
    int  flags, width, len, type;
    char fmt[32];

    /* TEST "%d" */
    for (flags=0; flags<SIZEOF(str_flags); flags++)
    {
        for (width=0; width<SIZEOF(str_width); width++)
        {
            for (len=0; len<SIZEOF(str_len); len++)
            {
                for (type=0; type<SIZEOF(str_type); type++)
                {
                    strcpy(fmt, "%");
                    strcat(fmt, str_flags[flags]);
                    strcat(fmt, str_width[width]);
                    strcat(fmt, str_len[len]);
                    strcat(fmt, str_type[type]);

                    switch (len) {
                    case 1:
                        if (width == 2)
                            test(buf1, buf2, fmt, 16, (short)x);
                        else
                            test(buf1, buf2, fmt, (short)x);
                        break;
                    case 2:
                        if (width == 2)
                            test(buf1, buf2, fmt, 16, (long)x);
                        else
                            test(buf1, buf2, fmt, (long)x);
                        break;
                    case 3:
                    case 6:
                        if (width == 2)
                            test(buf1, buf2, fmt, 16, (__int64)x);
                        else
                            test(buf1, buf2, fmt, (__int64)x);
                        break;
                    case 4:
                        if (width == 2)
                            test(buf1, buf2, fmt, 16, (size_t)x);
                        else
                            test(buf1, buf2, fmt, (size_t)x);
                        break;
                    case 5:
                        if (width == 2)
                            test(buf1, buf2, fmt, 16, (__int32)x);
                        else
                            test(buf1, buf2, fmt, (__int32)x);
                        break;

                    default:
                        if (width == 2)
                            test(buf1, buf2, fmt, 16, (int)x);
                        else
                            test(buf1, buf2, fmt, (int)x);
                        break;
                    }
                }
            }
        }
    }
}

void run_s_test(const char *str)
{
    const char *str_flags[] = {
        "", "-", "+", " ", "#", "0",
    };
    const char *str_width[] = {
        "", "20", "*",
    };
    const char *str_len[] = {
        "", "h", "l", "w"
    };
    char buf1[256];
    char buf2[256];
    int  flags, width, len;
    char fmt[32];
    wchar_t *wstr;

    if (str == NULL)
        wstr = NULL;
    else {
        wstr = (wchar_t *)alloca((strlen(str)+1)*sizeof(wchar_t));
        len = 0;
        do wstr[len] = str[len]; while (str[len++] != '\0');
    }

    /* TEST "%s" */
    for (flags=0; flags<SIZEOF(str_flags); flags++)
    {
        for (width=0; width<SIZEOF(str_width); width++)
        {
            for (len=0; len<SIZEOF(str_len); len++)
            {
                strcpy(fmt, "%");
                strcat(fmt, str_flags[flags]);
                strcat(fmt, str_width[width]);
                strcat(fmt, str_len[len]);
                strcat(fmt, "s");

                if (len == 2 || len == 3) {
                    if (width == 2)
                        test(buf1, buf2, fmt, 16, wstr);
                    else
                        test(buf1, buf2, fmt, wstr);
                } else {
                    if (width == 2)
                        test(buf1, buf2, fmt, 16, str);
                    else
                        test(buf1, buf2, fmt, str);
                }
            }
        }
    }

}

void run_S_test(const char *str)
{
    const char *str_flags[] = {
        "", "-", "+", " ", "#", "0",
    };
    const char *str_width[] = {
        "", "20", "*",
    };
    const char *str_len[] = {
        "", "h", "l", "w"
    };
    char buf1[256];
    char buf2[256];
    int  flags, width, len;
    char fmt[32];
    wchar_t *wstr;

    if (str == NULL)
        wstr = NULL;
    else {
        wstr = (wchar_t *)alloca((strlen(str)+1)*sizeof(wchar_t));
        len = 0;
        do wstr[len] = str[len]; while (str[len++] != '\0');
    }

    /* TEST "%s" */
    for (flags=0; flags<SIZEOF(str_flags); flags++)
    {
        for (width=0; width<SIZEOF(str_width); width++)
        {
            for (len=0; len<SIZEOF(str_len); len++)
            {
                strcpy(fmt, "%");
                strcat(fmt, str_flags[flags]);
                strcat(fmt, str_width[width]);
                strcat(fmt, str_len[len]);
                strcat(fmt, "S");

                if (len == 0 || len == 2 || len == 3) {
                    if (width == 2)
                        test(buf1, buf2, fmt, 16, wstr);
                    else
                        test(buf1, buf2, fmt, wstr);
                } else {
                    if (width == 2)
                        test(buf1, buf2, fmt, 16, str);
                    else
                        test(buf1, buf2, fmt, str);
                }
            }
        }
    }

}

void run_fp_test(long double x)
{
    const char *str_flags[] = {
        "", "-", "+", " ", "#", "0",
    };
    const char *str_width[] = {
        "", "12.", "12.12", "*.", "*.*",
    };
    const char *str_len[] = {
        "", "l", "L"
    };
    const char *str_type[] = {
        "e", "E", "f", "g", "G"
    };

    char buf1[256];
    char buf2[256];
    int  flags, width, len, type;
    char fmt[32];

    /* TEST "%d" */
    for (flags=0; flags<SIZEOF(str_flags); flags++)
    {
        for (width=0; width<SIZEOF(str_width); width++)
        {
            for (len=0; len<SIZEOF(str_len); len++)
            {
                for (type=0; type<SIZEOF(str_type); type++)
                {
                    strcpy(fmt, "%");
                    strcat(fmt, str_flags[flags]);
                    strcat(fmt, str_width[width]);
                    strcat(fmt, str_len[len]);
                    strcat(fmt, str_type[type]);

                    switch (width) {
                    case 0:
                    case 1:
                    case 2:
                        switch (len) {
                        case 2:
                            test(buf1, buf2, fmt, (long double)x);
                            break;
                        default:
                            test(buf1, buf2, fmt, (double)x);
                            break;
                        }
                        break;
                    case 3:
                        switch (len) {
                        case 2:
                            test(buf1, buf2, fmt, 8, (long double)x);
                            break;
                        default:
                            test(buf1, buf2, fmt, 8, (double)x);
                            break;
                        }
                        break;
                    case 4:
                        switch (len) {
                        case 2:
                            test(buf1, buf2, fmt, 8, 8, (long double)x);
                            break;
                        default:
                            test(buf1, buf2, fmt, 8, 8, (double)x);
                            break;
                        }
                        break;
                    }
                }
            }
        }
    }
}

int main()
{
    /* BYTE range */
    run_int_test(INT64_K(123));
    run_int_test(INT64_K(0));
    run_int_test(INT64_K(-123));

    /* WORD range */
    run_int_test(INT64_K(12345));
    run_int_test(INT64_K(-12345));

    /* DWORD range */
    run_int_test(INT64_K(1234567900));
    run_int_test(INT64_K(-12345678900));

    /* QWORD range */
    run_int_test(INT64_K(1234567900000));
    run_int_test(INT64_K(-12345678900000));

    /* string tests */
    run_s_test("hello");
    run_s_test(NULL);
    run_S_test("hello");
    run_S_test(NULL);

    /* FP test 1: print zero */
    run_fp_test(0.);

    /* FP test 2: no significant digits after dot */
    run_fp_test(123.);
    run_fp_test(-123.);

    /* FP test 3: some digits after dot */
    run_fp_test(123.456);
    run_fp_test(-123.456);

    /* FP test 4: same but grew the number */
    run_fp_test(112233.445566778899);
    run_fp_test(-112233.445566778899);

    /* FP test 5: test consistency of SCI notation */
    run_fp_test(-0.0004);
    run_fp_test(-0.000456);
    run_fp_test(-0.000445566778899);

    return 0;
}
